#![no_std]
#![no_main]

mod args;
mod caps;
mod link;

use args::Args;
use libc::{c_int, c_void, unlink, EXIT_FAILURE, EXIT_SUCCESS};
use libc_print::std_name::eprintln;

#[no_mangle]
unsafe extern "C" fn main(argc: c_int, argv: *const *const u8) -> c_int {
    match argc {
        3 => (),
        1 => return EXIT_SUCCESS,
        _ => {
            eprintln!("Usage: setcap capabilities filename");
            return EXIT_FAILURE;
        }
    }

    let arg = match Args::new(argc as isize, argv) {
        Ok(arg) => arg,
        Err((valid_up_to, err)) => {
            match err.error_len() {
                Some(err_len) => eprintln!(
                    "invalid utf-8 sequence of {} bytes from index {}",
                    err_len, valid_up_to
                ),
                None => eprintln!("incomplete utf-8 byte sequence from index {}", valid_up_to),
            };
            return EXIT_FAILURE;
        }
    };

    let my_caps = caps::cap_get_proc();

    if my_caps.is_null() {
        eprintln!("cap_get_proc");
        return EXIT_FAILURE;
    }

    let target_caps = caps::cap_from_text(arg.get_cstr_by_index(1).as_ptr());

    if target_caps.is_null() {
        eprintln!(
            "cap_from_text: failed to parse \"{}\"",
            arg.get_str_by_index(1),
        );
        caps::cap_free(my_caps as *mut c_void);
        return EXIT_FAILURE;
    }

    let ret = caps::cap_set_nsowner(target_caps, 0);
    if ret != 0 {
        eprintln!("cap_set_nsowner: {}", ret);
        caps::cap_free(my_caps as *mut c_void);
        caps::cap_free(target_caps as *mut c_void);
        return EXIT_FAILURE;
    }

    let flag: caps::cap_value_t = caps::Value::CAP_SETFCAP.into();
    let ret = caps::cap_set_flag(
        my_caps,
        caps::Flag::CAP_EFFECTIVE.into(),
        1,
        &flag,
        caps::FlagValue::CAP_SET.into(),
    );

    if ret != 0 {
        eprintln!("cap_set_flag(CAP_SETFCAP): {}", ret);
        caps::cap_free(my_caps as *mut c_void);
        caps::cap_free(target_caps as *mut c_void);
        return EXIT_FAILURE;
    }

    let ret = caps::cap_set_proc(my_caps);
    if ret != 0 {
        eprintln!("cap_set_proc: {}", ret);
        caps::cap_free(my_caps as *mut c_void);
        caps::cap_free(target_caps as *mut c_void);
        return EXIT_FAILURE;
    }

    caps::cap_free(my_caps as *mut c_void);

    let ret = caps::cap_set_file(arg.get_cstr_by_index(2).as_ptr(), target_caps);
    if ret != 0 {
        eprintln!("cap_set_file: {}", ret);
        caps::cap_free(target_caps as *mut c_void);
        return EXIT_FAILURE;
    }

    caps::cap_free(target_caps as *mut c_void);

    _ = match arg.get_cstr_by_index(0).to_bytes() {
        [b'/', b'!', ..] => unlink(arg.get_cstr_by_index(0).as_ptr()),
        _ => 0,
    };

    EXIT_SUCCESS
}
